from numpy.testing import assert_equal, assert_array_almost_equal, assert_array_equal
from ..sequence_plots import *


def test_compute_entropy():
    # entropy values are computed here: http://www.shannonentropy.netmark.pl/calculate
    X = [list('123123123123121212'), list('absfdvebavdgabsavgasb'), list('1111111111111111111111'),
         list('123456789'), list('00000000000000'),
         list('123456789gsadfvop8u4w3hji0as7drfpiw3o4htfya’0sdfgphi0arepgiaudfgt’8piuhagv')]
    entropies = [1.54198, 2.81520, .0, 3.16993, .0, 4.58138]
    assert_array_almost_equal(entropies, SequencePlots().compute_entropy_(X), 5, 'Entropy of X is wrong')


def test_string_arr_to_int_matrix_():
    X = np.array(['AAABB', 'AAABBBB', 'CDACCBBA', 'CDACCBB'])
    X_ascii = [[65, 65, 65, 66, 66, 0, 0, 0],  # 0 padding
               [65, 65, 65, 66, 66, 66, 66, 0],
               [67, 68, 65, 67, 67, 66, 66, 65],
               [67, 68, 65, 67, 67, 66, 66, 0]]
    assert_equal(SequencePlots().string_arr_to_int_matrix_(X), X_ascii)


def test_find_alphabet_():
    X = ['aabbcder', 'saefs']
    alphabet = list('abcdefrs')
    assert_array_equal(np.sort(SequencePlots().find_alphabet_(X)), np.sort(alphabet))



from numpy.testing import assert_equal

from ..twitter_sequencer import *


# TODO fit transform from input_file


def dummy_data_gen_():
    X = []
    in_reply_to_user_id = [0, None, 2, 3, None, None]
    retweeted_status = [0, 1, 2, None, 5, None]
    uid = [111, 111, 111, 222, 222, 222]
    entities = [{}, {'url': '123'}, {'hashtags': '123'}, {'user_mentions': '123'}, {'extended_entities': '123'},
                {'user_mentions': '123', 'url': '123'}]
    for i in range(len(in_reply_to_user_id)):
        tweet = {'in_reply_to_user_id': retweeted_status[i], 'retweeted_status': {'id': retweeted_status[i]},
                 'user': {'id': uid[i]}, 'entities': entities[i]}
        X.append(tweet)
    Y_b3t = [[111, TWEET + RETWEET + RETWEET], [222, TWEET + RETWEET + TWEET]]
    Y_b3c = [[111, NONE + ENTITY + ENTITY], [222, ENTITY + ENTITY + MIXED]]
    Y_b6c = [[111, NONE + URL + HASHTAG], [222, MENTION + UNKNOWN + MIXED]]
    return X, Y_b3t, Y_b3c, Y_b6c


X, Y_b3t, Y_b3c, Y_b6c = dummy_data_gen_()


def test_b3_type():
    estimator = TwitterDDNASequencer(alphabet='b3_type')
    X_transformed = estimator.fit_transform(X)
    assert_equal(X_transformed, Y_b3t)


def test_b3_content():
    estimator = TwitterDDNASequencer(alphabet='b3_content')
    X_transformed = estimator.fit_transform(X)
    assert_equal(X_transformed, Y_b3c)


def test_b6_content():
    estimator = TwitterDDNASequencer(alphabet='b6_content')
    X_transformed = estimator.fit_transform(X)
    assert_equal(X_transformed, Y_b6c)



"""
This is a module to be used as a reference for building other modules
"""
import warnings
from glcr import longest_common_subsequence
from multiprocessing import Process

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array

from .verbosity import Verbosity


# noinspection PyIncorrectDocstring
class LongestCommonSubsequence(BaseEstimator):
    """ The Digital DNA Python implementation.

    Parameters
    ----------
    in_path : str, optional
        The name with absolute path of a file containing the sequences you want to process.
        The input file must be a txt file and the first row must contain the number of sequences to read.
        Default: ''

    out_path : str, optional
        The output file name with absolute path of the file where the algorithm will save result in case of
        verbosity equals to Verbosity.FILE or Verbosity.FILE_EXTENDED
        Default: '/tmp/glcr_cache'

    overwrite : boolean, optional
        It must be False to use the LCS files produced in a previous fit call, in this case the file names are
        the ones specified in the out_path parameter. If True, recomputes the LCS files.
        Default: False

    window : str, optional
        The size of the window used to compute the cutting threshold between bot and not bot. The cutting point is
        computed by smoothing the curve, deriving the result and taking the first (l.h.s.) local maxima. The window
        parameter influences both smoothing and finding the local maxima.
        It must be 0 < window < n_accounts. If
        - window >= 1 , window is the number of elements
        - 0 < window < 1, window is the percentage of the number of elements
        Default: 10

    verbosity : str, optional
        The verbosity parameter is used to specify whether to save results to files or not. It must be:
        - TEST              does not write anything, used for benchmarking
        - MEMORY_ONLY       retrieves only the couples (sequence length, # of accounts), used for plots
        - FILE              produces 2 files, a file named self.out_path + '.mat' where each row contains:
                            - sequence length,
                            - # of accounts
                            - range of indexes (begin and end).
                            And a file named self.out_path + '.gsa' where each row contains the identifier of the
                            sequence.
        - FILE_EXTENDED     as FILE but the self.in_path + '.mat' file contains also the column of the common
                            subsequence

    References
    ----------
    S. Cresci, R. D. Pietro, M. Petrocchi, A. Spognardi and M. Tesconi,
    "Social Fingerprinting: Detection of Spambot Groups Through DNA-Inspired Behavioral Modeling",
    IEEE Transactions on Dependable and Secure Computing, vol. 15, no. 4, pp. 561-576, 1 July-Aug. 2018,
    https://ieeexplore.ieee.org/document/7876716

    S. Cresci, R. di Pietro, M. Petrocchi, A. Spognardi and M. Tesconi,
    "Exploiting Digital DNA for the Analysis of Similarities in Twitter Behaviours",
    2017 IEEE International Conference on Data Science and Advanced Analytics (DSAA),
    Tokyo, 2017, pp. 686-695, https://ieeexplore.ieee.org/document/8259831

    M. Arnold, E. Ohlebusch, "Linear Time Algorithms for Generalizations of the Longest Common Substring Problem",
    Algorithmica, vol 60, pp. 806-818, 4 August 2011, https://link.springer.com/article/10.1007/s00453-009-9369-1
    """
    def __init__(self, in_path='', out_path='/tmp/glcr_cache', overwrite=False, window=10, verbosity=Verbosity.FILE):
        self.in_path = in_path
        self.out_path = out_path
        self.overwrite = overwrite
        self.window = window
        if self.window < 2:
            raise ValueError('window parameter cannot be less than 2.')
        self.verbosity = verbosity

    def fit(self, X, y=None):
        """Computes the longest common subsequence

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Attributes
        ----------
        lcs_index_ : pandas dataframe, shape (distinct couples (lcs, n_of_accounts), 2), default=None
            The dataframe containing the distict couples lcs, n_of_accounts). Only if:
            verbosity == Verbosity.MEMORY_ONLY

        Returns
        -------
        self : object
            Returns self.
        """
        warnings.simplefilter(action='ignore', category=FutureWarning)
        self.gsa_ = '.gsa'
        self.mat_ = '.mat'

        if self.in_path == '':
            X = self._unicode_to_ascii(X)
            X = check_array(X, ensure_2d=False)
        print("fitting...")
        if self.verbosity == Verbosity.MEMORY_ONLY:
            self.lcs_ = longest_common_subsequence(self._unicode_to_ascii(X), self.in_path, self.out_path,
                                                   self.verbosity)
        elif self.verbosity > Verbosity.MEMORY_ONLY:
            if self.in_path == '' or self.overwrite:
                p = Process(target=longest_common_subsequence, args=(X, self.in_path, self.out_path, self.verbosity))
                p.start()
                p.join()
        return self

    def predict(self, X=None):
        """ (EXPERIMENTAL) Predict the labels (True bot, False Not Bot) of X according to lcs and window parameter:.
        If X is None, returns the same as fit_predict(X_train).
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features), default=None
            The query sample or samples to identify bot groups. If None,
            makes prediction on the training data.
        Returns
        -------
        y : array, shape (n_samples,)
            Returns True for bots and False for real timeline.
        """
        y = None
        if self.verbosity > Verbosity.MEMORY_ONLY:
            y = np.full(len(X), False)
            mat_ix = set()
            print('finding cut...')
            lengths = pd.read_csv(self.out_path+self.mat_, usecols=['length']).length.values
            self.cut_ = self._decision_function(lengths)
            del lengths
            print('predicting...')

            class BreakIt(Exception):
                pass
            try:
                for df in pd.read_csv(self.out_path + self.mat_, chunksize=500000):
                    for ix, row in df.iterrows():
                        curr_set = mat_ix.union(set(range(int(row.begin), int(row.end) + 1)))
                        if row.num_texts >= self.cut_:  # not bot, exit
                            raise BreakIt
                        mat_ix = curr_set
            except BreakIt:
                pass
            ix_size = len(mat_ix)
            ix_count = 0
            bot_ix = np.empty(ix_size, dtype=int)
            try:
                for df in pd.read_csv(self.out_path + self.gsa_, header=0, usecols=['wordindex'], squeeze=True,
                                      chunksize=500000):
                    for ix, wordindex in df.items():
                        if ix in mat_ix:
                            # TODO delete the -1 when fix the gsa file
                            bot_ix[ix_count] = wordindex - 1
                            ix_count += 1
                            if ix_count == ix_size:
                                raise BreakIt
                            mat_ix.remove(ix)
            except BreakIt:
                y[bot_ix] = True
        else:
            warnings.warn("Cannot predict with verbosity level lower than FILE", Warning)
        print('done.')
        return y

    def fit_predict(self, X, y=None):
        """"Fits the model to the training set X and returns the labels
        (True for bot, False for non bot) on the training set according to the window parameter.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features), default=None
            The query sample or samples to compute the Local Outlier Factor
            w.r.t. to the training samples.

        Attributes
        ----------
        lcs_index_ : pandas dataframe, shape (distinct couples (lcs, n_of_accounts), 2), default=None
            The dataframe containing the distict couples lcs, n_of_accounts)

        Returns
        -------
        y : array, shape (n_samples,)
            Returns True for bots and False for real timeline.
        """
        return self.fit(X).predict(X)

    def plot_LCS(self):
        """"Plots the longest common subsequence curve as (number of accounts, sequence length)

        Attributes
        ----------
        lcs_index_ : pandas dataframe, shape (distinct couples (lcs, n_of_accounts), 2), default=None
            The dataframe containing the distict couples lcs, n_of_accounts)

        Returns
        -------
        self : returns an instance of self.
        """
        plt.xlabel('# of accounts')
        plt.ylabel('LCS')
        if self.verbosity > Verbosity.MEMORY_ONLY and not hasattr(self, 'lcs_'):
            self.lcs_ = pd.read_csv(self.out_path + self.mat_, usecols=['length', 'num_texts']) \
                .drop_duplicates().reset_index().drop(['index'], axis=1)
        plt.plot(self.lcs_.num_texts, self.lcs_.length, marker='x')
        if hasattr(self, 'cut_'):
            plt.plot([self.cut_, self.cut_], [0, max(self.lcs_.length)], linestyle='--')
        plt.show()
        return self

    def plot_LCS_log(self):
        """"Plots the longest common subsequence curve as (log(number of accounts), log(sequence length))

        Attributes
        ----------
        lcs_index_ : pandas dataframe, shape (distinct couples (lcs, n_of_accounts), 2), default=None
            The dataframe containing the distict couples lcs, n_of_accounts)

        Returns
        -------
        self : returns an instance of self.
        """
        plt.xlabel('log(# of accounts)')
        plt.ylabel('log(LCS)')
        if self.verbosity > Verbosity.MEMORY_ONLY and not hasattr(self, 'lcs_'):
            self.lcs_ = pd.read_csv(self.out_path + self.mat_, usecols=['length', 'num_texts']) \
                .drop_duplicates().reset_index().drop(['index'], axis=1)
        plt.loglog(self.lcs_.num_texts, self.lcs_.length, marker='x')
        if hasattr(self, 'cut_'):
            plt.plot([self.cut_, self.cut_], [0, max(self.lcs_.length)], linestyle='--')
        plt.show()
        return self

    def _decision_function(self, X):
        ''' Finds the first relative maximum on the smoothed LCS vector'''
        length = len(X)
        w_size = self.window
        X_avg_diff = np.diff(self._running_mean(X, w_size))
        i = 0
        max_index = length - w_size
        for i in range(max_index):
            if X_avg_diff[i] < np.mean(X_avg_diff[i:i+w_size]):
                break
        t = np.argmax(X_avg_diff[i:])
        return t + i - 1

    def _running_mean(self, x, winsize):
        for i in range(len(x)):
            x[i] = x[max(0, i):min(i + winsize, len(x))].mean()
        return x

    def _unicode_to_ascii(self, array):
        def f(item):
            return str(item) + '\0'

        def v(x):
            return np.vectorize(f)(x)

        return v(array).astype('S')



"""
Digital DNA Sequencer for Twitter
"""
import json
import numpy as np

ENTITY = 'E'
HASHTAG = 'T'
MEDIA = 'G'
MENTION = 'C'
NONE = 'N'
MIXED = 'X'
REPLY = 'T'
RETWEET = 'C'
TWEET = 'A'
URL = 'A'
UNKNOWN = 'U'


class TwitterDDNASequencer():  # BaseEstimator, TransformerMixin):
    """ Twitter Digital DNA Sequencer.
    Compute sequences of digital DNA from twitter timelines (check out
    https://developer.twitter.com/en/docs/tweets/timelines/api-reference/get-statuses-user_timeline.html)

    Parameters
    ----------
    alphabet :  string or callable, default ‘b3_type’
        mapping between the column value and the corresponding base.
        If alphabet_ is a callable function, it is called on each
        pair of instances (rows) and the resulting value recorded.
        The callable should take two arrays as input and return
        one value indicating the distance between them.
        Prebuild alphabets are the following:
        - 'b3_type',  where the correspondence  is
                - 'A' for tweet
                - 'C' for reply
                - 'T' for retweet
        - 'b3_content', where the correspondence  is
                - 'N' tweet contains no entities (plain text)
                - 'E' tweet contains entities of one type
                - 'X' tweet contains entities of mixed types
        - 'b6_content', where the correspondence  is
                - 'N' tweet contains no entities (plain text)
                - 'U' tweet contains one or more URLs
                - 'H' tweet contains one or more hashtags
                - 'M' tweet contains one or more mentions
                - 'D' tweet contains one or more medias
                - 'X' tweet contains entities of mixed types

    Attributes
    ----------
    input_shape : tuple
        The shape the data passed to :meth:`fit`


    References
    ----------
    S. Cresci, R. D. Pietro, M. Petrocchi, A. Spognardi and M. Tesconi,
    "Social Fingerprinting: Detection of Spambot Groups Through DNA-Inspired Behavioral Modeling",
    IEEE Transactions on Dependable and Secure Computing, vol. 15, no. 4, pp. 561-576, 1 July-Aug. 2018,
    https://ieeexplore.ieee.org/document/7876716

    S. Cresci, R. di Pietro, M. Petrocchi, A. Spognardi and M. Tesconi,
    "Exploiting Digital DNA for the Analysis of Similarities in Twitter Behaviours",
    2017 IEEE International Conference on Data Science and Advanced Analytics (DSAA),
    Tokyo, 2017, pp. 686-695, https://ieeexplore.ieee.org/document/8259831
    """

    def __init__(self, alphabet='b3_type', input_file=''):
        import warnings
        warnings.simplefilter(action='ignore', category=FutureWarning)
        self.alphabet = alphabet
        self.input_file = input_file

    def fit(self, X=None, y=None):
        """Simply assigns the right alphabet mapper function to remap_

        Parameters
        ----------
        X : None
            The pipeline API requires this parameter.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Attributes
        ----------
        remap_ : function that takes a tweet with the same format retrieved from
                GET statuses/user_timeline call and retrieves the corresponding character
                given the alphabet parameter.

        Returns
        -------
        self : object
            Returns an instance of self.
        """
        if self.alphabet == 'b3_type':
            self.remap_ = self.tweet2char_b3_type_
        elif self.alphabet == 'b3_content':
            self.remap_ = self.tweet2char_b3_content_
        elif self.alphabet == 'b6_content':
            self.remap_ = self.tweet2char_b6_content_
        else:
            self.remap_ = None

        return self

    def transform(self, X=None):
        """ The function that transform the array of timelines to digital dna sequences
            given the alphabet.

        Parameters
        ----------
        X : array-like of shape = [# of tweets, 1]
            The input samples, each sample is a python dict of a tweet as retrieved
            from twitter user timelines API (check out
            https://developer.twitter.com/en/docs/tweets/timelines/api-reference/get-statuses-user_timeline.html)

        Returns
        -------
        X_transformed : array of string of shape = [# of users, 1]
            The array containing the digital dna sequences
            """
        # Check is fit had been called
        # check_is_fitted(self, ['input_shape_'])

        # Input validation
        # X = check_array(X)

        # Check that the input is of the same shape as the one passed
        # during fit.
        # if X.shape != self.input_shape_:
        #    raise ValueError('Shape of input is different from what was seen'
        #                     'in `fit`')
        if self.input_file != '':
            f = open(self.input_file, "r")
            X = json.loads(f.read())
        elif X == None:
            raise ValueError('X cannot be None if input_file is not specified')

        ddna_size = 0
        res = {}
        for row in X:
            uid = self.nested_get_(row, 'user.id')
            code = self.remap_(row)
            if uid in res:
                res[uid] += code
            else:
                res[uid] = code
            ddna_size = max(ddna_size, len(res[uid]))
        # dtype = [('uid', 'i8'), ('ddna', 'U' + str(ddna_size))]
        res = np.array(list(res.items()))  # , dtype=dtype)
        return res

    def fit_transform(self, X=None, y=None):
        """ Fit and transform

        Parameters
        ----------
        X : array-like of shape = [# of tweets, 1]
            The input samples, each sample is a python dict of a tweet as retrieved
            from twitter user timelines API (check out
            https://developer.twitter.com/en/docs/tweets/timelines/api-reference/get-statuses-user_timeline.html)

        y: ignored parameter needed to mantain a standard pattern

        Returns
        -------
        X_transformed : array of shape = [n_samples, 2]
            The resulting array where the first column is the user id and the second is the translated sequence
        """
        # Input validation
        # X = check_array(X)
        # self.input_shape_ = X.shape

        # Check that the input is of the same shape as the one passed
        # during fit.
        # if X.shape != self.input_shape_:
        #    raise ValueError('Shape of input is different from what was seen'
        #                     'in `fit`')
        return self.fit(X, y).transform(X)

    def nested_get_(self, dct, keys):
        for key in keys.split('.'):
            try:
                dct = dct[key]
            except KeyError:
                return None
        return dct

    def tweet2char_b3_type_(self, tweet):
        reply_id = tweet['in_reply_to_user_id']
        is_reply = reply_id != 0 and reply_id is not None
        retweet_id = self.nested_get_(tweet, 'retweeted_status.id')
        is_retweet = retweet_id != 0 and retweet_id is not None

        if not is_reply and not is_retweet:
            return TWEET
        elif is_reply and not is_retweet:
            return REPLY
        return RETWEET

    def tweet2char_b3_content_(self, tweet):
        # sums not empty lists within entities dict
        n_entities = sum([1 for k, v in tweet['entities'].items() if v])

        if n_entities == 0:
            return NONE
        elif n_entities > 1:
            return MIXED
        return ENTITY

    def tweet2char_b6_content_(self, tweet):
        # sums not empty lists within entities dict
        n_entities = sum([1 for k, v in tweet['entities'].items() if v])

        if n_entities == 0:
            return NONE
        elif n_entities > 1:
            return MIXED
        else:
            type = [k for k, v in tweet['entities'].items() if v][0]
            if type == 'url':
                return URL
            elif type == 'hashtags':
                return HASHTAG
            elif type == 'user_mentions':
                return MENTION
            elif 'extended_entities' in tweet:
                return MEDIA
            return UNKNOWN

    def get_params(self):
        return {'alphabet': self.alphabet,
                'input_file': self.input_file}

    def set_params(self, alphabet, input_file):
        self.alphabet = alphabet
        self.input_file = input_file
